import os
import tempfile
from pathlib import Path
from unittest.mock import ANY, Mock, PropertyMock, patch

import click
import pytest

from energinetml.cli.utils import (
    parse_input_path,
    parse_input_project_name,
    parse_input_resource_group,
    parse_input_subscription_id,
    parse_input_workspace_name,
)

NAME = "PROJECT-NAME"
SUBSCRIPTION_NAME = "MY-SUBSCRIPTION"
SUBSCRIPTION_ID = "SUBSCRIPTION-ID"
RESOURCE_GROUP = "RESOURCE-GROUP"
WORKSPACE_NAME = "WORKSPACE-NAME"
VNET = "VNET"
SUBNET = "SUBNET"


# -- _parse_input_path() Tests -----------------------------------------------


@patch("energinetml.cli.project.init.click.prompt")
def test__parse_input_path__value_is_none__should_prompt_for_path(prompt_mock):
    """
    :param Mock prompt_mock:
    """

    # Act
    parse_input_path([])(ctx=None, param=None, value=None)

    # Assert
    prompt_mock.assert_called_once_with(
        text="Enter project location", default=ANY, type=ANY
    )


@patch("energinetml.cli.project.init.click.echo")
def test__parse_input_path__path_is_a_file__should_abort_with_error(echo_mock):
    """
    :param Mock echo_mock:
    """
    with tempfile.TemporaryDirectory() as path:
        fp = os.path.join(path, "somefile.txt")

        Path(fp).touch()

        # Act + Assert
        with pytest.raises(click.Abort):
            parse_input_path([])(ctx=None, param=None, value=fp)


@patch("energinetml.cli.project.init.click.echo")
@patch("energinetml.cli.project.init.click.confirm")
def test__parse_input_path__project_files_already_exists_in_folder__should_prompt_to_override(  # noqa: E501
    confirm_mock, echo_mock
):
    """
    :param Mock confirm_mock:
    :param Mock echo_mock:
    """
    confirm_mock.side_effect = (True, False)

    with tempfile.TemporaryDirectory() as path:
        fp1 = os.path.join(path, "project.json")
        fp2 = os.path.join(path, "requirements.txt")

        Path(fp1).touch()
        Path(fp2).touch()

        # Act
        with pytest.raises(click.Abort):
            project_files = ["project.json", "requirements.txt"]
            parse_input_path(project_files)(ctx=None, param=None, value=path)

        # Assert
        echo_mock.assert_any_call("File already exists: %s" % fp1)
        echo_mock.assert_any_call("File already exists: %s" % fp2)

        confirm_mock.assert_any_call("Really override existing project.json?")
        confirm_mock.assert_any_call("Really override existing requirements.txt?")


# -- _parse_input_project_name() Tests ---------------------------------------


@patch("energinetml.cli.project.init.click.prompt")
def test__parse_input_project_name__value_is_none__should_prompt_for_name(
    prompt_mock,
):
    """
    :param Mock prompt_mock:
    """
    params = {"path": "project-path"}
    prompt_mock.return_value = NAME

    # Act
    value = parse_input_project_name()(ctx=Mock(params=params), param=None, value=None)

    # Assert
    prompt_mock.assert_called_once_with(
        text="Please enter a project name", default="project-path", type=str
    )

    assert value == NAME


@patch("energinetml.cli.project.init.click.prompt")
def test__parse_input_project_name__value_is_provided__should_return_value(
    prompt_mock,
):
    """
    :param Mock prompt_mock:
    """

    # Act
    value = parse_input_project_name()(ctx=None, param=None, value=NAME)

    # Assert
    assert value == NAME
    prompt_mock.assert_not_called()


# -- _parse_input_subscription_id() Tests ------------------------------------


@patch("energinetml.cli.utils.projects.backend")
@patch("energinetml.cli.project.init.click.prompt")
def test__parse_input_subscription_id__value_not_known_subscription__should_prompt_for_subscription_name(  # noqa: E501
    prompt_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    """
    prompt_mock.return_value = SUBSCRIPTION_NAME

    subscription = Mock(subscription_id=SUBSCRIPTION_ID, display_name=SUBSCRIPTION_NAME)

    backend_mock.get_available_subscriptions.return_value = [subscription]

    # Act
    value = parse_input_subscription_id()(ctx=None, param=None, value="foobar")

    # Assert
    prompt_mock.assert_called_once_with(
        text="Please enter Azure Subscription", type=ANY
    )

    assert value == SUBSCRIPTION_ID


# -- _parse_input_resource_group() Tests -------------------------------------


@patch("energinetml.cli.utils.projects.backend")
@patch("energinetml.cli.project.init.click.prompt")
def test__parse_input_resource_group__value_not_known_resource_group__should_prompt_for_resource_group(  # noqa: E501
    prompt_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    """
    ctx = Mock(params={"subscription_id": SUBSCRIPTION_ID})
    prompt_mock.return_value = RESOURCE_GROUP

    # Workaround: Can not mock property "name" of Mock object
    resource_group = Mock()
    p = PropertyMock(return_value=RESOURCE_GROUP)
    type(resource_group).name = p

    backend_mock.get_available_resource_groups.return_value = [resource_group]

    # Act
    value = parse_input_resource_group()(ctx=ctx, param=None, value="foobar")

    # Assert
    prompt_mock.assert_called_once_with(
        text="Please enter Azure Resource Group", default="", type=ANY
    )

    assert value == RESOURCE_GROUP


# -- _parse_input_workspace_name() Tests -------------------------------------


def test__parse_input_workspace_name__no_subscription_id_available__should_raise_runtime_error():  # noqa: E501
    ctx = Mock(params={"resource_group": RESOURCE_GROUP})

    with pytest.raises(RuntimeError):
        parse_input_workspace_name()(ctx=ctx, param=None, value=None)


def test__parse_input_workspace_name__no_resource_group_available__should_raise_runtime_error():  # noqa: E501
    ctx = Mock(params={"subscription_id": SUBSCRIPTION_ID})

    with pytest.raises(RuntimeError):
        parse_input_workspace_name()(ctx=ctx, param=None, value=None)


@patch("energinetml.cli.utils.projects.backend")
@patch("energinetml.cli.project.init.click.prompt")
def test__parse_input_workspace_name__no_workspaces_exists_in_resource_group__should_prompt_for_workspace(  # noqa: E501
    prompt_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    """
    ctx = Mock(
        params={"subscription_id": SUBSCRIPTION_ID, "resource_group": RESOURCE_GROUP}
    )

    backend_mock.get_available_workspace_names.return_value = []
    prompt_mock.return_value = "workspace!"

    # Act
    value = parse_input_workspace_name()(ctx=ctx, param=None, value=None)

    # Assert
    backend_mock.get_available_workspace_names.assert_called_once_with(
        subscription_id=SUBSCRIPTION_ID, resource_group=RESOURCE_GROUP
    )

    prompt_mock.assert_called_once_with(
        text="Please enter AzureML Workspace name", default=None, type=ANY
    )

    assert value == "workspace!"


@patch("energinetml.cli.utils.projects.backend")
@patch("energinetml.cli.project.init.click.prompt")
def test__parse_input_workspace_name__no_workspaces_provided__should_prompt_for_workspace(  # noqa: E501
    prompt_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    """
    ctx = Mock(
        params={"subscription_id": SUBSCRIPTION_ID, "resource_group": RESOURCE_GROUP}
    )

    backend_mock.get_available_workspace_names.return_value = [WORKSPACE_NAME]
    prompt_mock.return_value = "workspace!"

    # Act
    value = parse_input_workspace_name()(ctx=ctx, param=None, value=None)

    # Assert
    prompt_mock.assert_called_once_with(
        text="Please enter AzureML Workspace name", default=WORKSPACE_NAME, type=ANY
    )

    assert value == "workspace!"



import json
from unittest.mock import ANY, MagicMock, Mock, PropertyMock, patch

from click.testing import CliRunner
from packaging.version import Version

from energinetml import PACKAGE_NAME
from energinetml.cli.model.submit import submit
from energinetml.core.model import Model
from tests.constants import COMPUTE_TARGET, VM_SIZE


class TestModelSubmit:
    def test__download_and_not_wait__should_abort(self, model_path):
        """
        :param str model_path:
        """
        runner = CliRunner()

        # Act
        result = runner.invoke(
            cli=submit, args=["--path", model_path, "--download", "--nowait"]
        )

        # Assert
        assert result.exit_code == 1
        assert result.output.startswith(
            "Can not use -d/--download together with --nowait"
        )

    @patch.object(Model, "requirements", new_callable=PropertyMock)
    def test__package_name_not_in_requirements__should_abort(
        self, model_requirements_mock, model_path
    ):
        """
        :param Mock model_requirements_mock:
        :param str model_path:
        """
        runner = CliRunner()

        model_requirements_mock.return_value = []

        # Act
        result = runner.invoke(cli=submit, args=["--path", model_path])

        # Assert
        assert result.exit_code == 1
        assert result.output.startswith(
            f"Could not find '{PACKAGE_NAME}' in the project's requirements.txt file"
        )

    @patch("energinetml.cli.model.submit.backend")
    @patch.object(Model, "requirements", new_callable=MagicMock)
    def test__required_sdk_version_less_than_current__should_echo_warning(
        self, model_requirements_mock, backend_mock, model_path, clusters_mock
    ):
        """
        :param Mock model_requirements_mock:
        :param Mock backend_mock:
        :param str model_path:
        """
        runner = CliRunner()

        model_requirements_mock.__contains__.return_value = True
        model_requirements_mock.get_version.return_value = Version("0.0.0")

        backend_mock.get_compute_clusters.return_value = clusters_mock

        # Act
        result = runner.invoke(cli=submit, args=["--path", model_path])

        # Assert
        assert result.exit_code == 0
        assert (
            (
                "WARNING: Your requirements.txt file contains a version of "
                "%s (0.0.0) which is older than your current installation"
            )
            % PACKAGE_NAME
        ) in result.output

    @patch("energinetml.cli.model.submit.backend")
    @patch.object(Model, "requirements", new_callable=MagicMock)
    def test__required_sdk_version_greater_than_current__should_echo_warning(
        self, model_requirements_mock, backend_mock, model_path, clusters_mock
    ):
        """
        :param Mock model_requirements_mock:
        :param Mock backend_mock:
        :param str model_path:
        """
        runner = CliRunner()

        model_requirements_mock.__contains__.return_value = True
        model_requirements_mock.get_version.return_value = Version("9999.9999.9999")

        backend_mock.get_compute_clusters.return_value = clusters_mock

        # Act
        result = runner.invoke(cli=submit, args=["--path", model_path])

        # Assert
        assert result.exit_code == 0
        assert (
            (
                "WARNING: Your requirements.txt file contains a version of "
                "%s (9999.9999.9999) which is newer than your current installation "
            )
            % PACKAGE_NAME
        ) in result.output

    @patch("energinetml.cli.model.submit.backend")
    @patch.object(Model, "requirements", new_callable=MagicMock)
    def test__should_submit_wait_and_download_files(
        self, model_requirements_mock, backend_mock, model_path, clusters_mock
    ):
        """
        :param Mock model_requirements_mock:
        :param Mock backend_mock:
        :param str model_path:
        """
        runner = CliRunner()

        model_requirements_mock.__contains__.return_value = True
        model_requirements_mock.get_version.return_value = Version("9999.9999.9999")

        context = Mock()

        backend_mock.submit_model.return_value = context
        backend_mock.get_compute_clusters.return_value = clusters_mock

        # Act
        result = runner.invoke(
            cli=submit,
            args=[
                "--path",
                model_path,
                "--wait",
                "--download",
                "parameter1",
                "parameter2",
            ],
        )

        # Assert
        assert result.exit_code == 0

        backend_mock.submit_model.assert_called_once_with(
            model=ANY, params=("parameter1", "parameter2")
        )

        context.wait_for_completion.assert_called_once()
        context.wait_for_completion.download_files()

    @patch("energinetml.cli.model.submit.backend")
    def test__verify_compute_cluster__should_prompt_user(
        self, backend_mock, model_path
    ):
        # Arrange
        cluster1 = Mock()
        cluster1.name = "cluster1"

        cluster2 = Mock()
        cluster2.name = "cluster2"
        cluster2.vm_size = VM_SIZE

        mock_clusters = [cluster1, cluster2]
        backend_mock.get_compute_clusters.return_value = mock_clusters

        backend_mock.get_workspace.return_value = Mock()
        backend_mock.submit_model.return_value = Mock()

        runner = CliRunner()

        # Act
        result = runner.invoke(
            cli=submit,
            args=[
                "--path",
                model_path,
            ],
            # Select cluster2 as the new cluster
            input="cluster2",
        )

        # Assert
        assert result.exit_code == 0
        assert (
            f"Could not find compute target '{COMPUTE_TARGET}' in "
            "the available compute targets: ['cluster1', 'cluster2']."
        ) in result.output

        # Check that cluster2 was selected.
        with open(f"{model_path}/{Model._CONFIG_FILE_NAME}", "r") as f:
            model_json = json.load(f)

        assert model_json["compute_target"] == "cluster2"

    @patch("energinetml.cli.model.submit.backend")
    def test__verify_compute_cluster__should_be_silent(self, backend_mock, model_path):
        # Arrange
        cluster1 = Mock()
        cluster1.name = COMPUTE_TARGET
        cluster1.vm_size = VM_SIZE

        cluster2 = Mock()
        cluster2.name = "cluster2"

        mock_clusters = [cluster1, cluster2]
        backend_mock.get_compute_clusters.return_value = mock_clusters
        backend_mock.get_workspace.return_value = Mock()
        backend_mock.submit_model.return_value = Mock()

        runner = CliRunner()

        # Act
        result = runner.invoke(
            cli=submit,
            args=[
                "--path",
                model_path,
            ],
        )

        # Assert
        assert result.exit_code == 0
        assert (
            f"Could not find compute target '{COMPUTE_TARGET}'"
        ) not in result.output

    @patch("energinetml.cli.model.submit.backend")
    def test__verify_vm_size__mismatching_vm_sizes_should_abort(
        self, backend_mock, model_path
    ):
        # Arrange
        cluster1 = Mock()
        cluster1.name = COMPUTE_TARGET
        cluster1.vm_size = "not the right vm size"
        mock_clusters = [cluster1]

        backend_mock.get_compute_clusters.return_value = mock_clusters
        backend_mock.get_workspace.return_value = Mock()
        backend_mock.submit_model.return_value = Mock()

        runner = CliRunner()

        # Act
        result = runner.invoke(
            cli=submit,
            args=[
                "--path",
                model_path,
            ],
        )

        # Assert
        assert result.exit_code == 1
        assert (
            "WARNING: The VM size (vm-size) specified in "
            f"{Model._CONFIG_FILE_NAME} does "
            "not match the actual VM size of the compute cluster "
            "(not the right vm size)."
        ) in result.output

    @patch("energinetml.cli.model.submit.backend")
    def test__verify_vm_size__matching_vm_sizes_should_be_silent(
        self, backend_mock, model_path
    ):
        # Arrange
        cluster1 = Mock()
        cluster1.name = COMPUTE_TARGET
        cluster1.vm_size = VM_SIZE
        mock_clusters = [cluster1]

        backend_mock.get_compute_clusters.return_value = mock_clusters
        backend_mock.get_workspace.return_value = Mock()
        backend_mock.submit_model.return_value = Mock()

        runner = CliRunner()

        # Act
        result = runner.invoke(
            cli=submit,
            args=[
                "--path",
                model_path,
            ],
        )

        # Assert
        assert result.exit_code == 0
        assert (
            f"WARNING: The VM size specified in {Model._CONFIG_FILE_NAME} does "
            "not match the actual VM size of the compute cluster."
        ) not in result.output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

import random
from typing import List
from unittest.mock import Mock

import azureml
import click
from azureml.core import Experiment

from energinetml.core.files import FileMatcher, temporary_folder
from energinetml.settings import PACKAGE_NAME, PACKAGE_VERSION, PYTHON_VERSION

from .backend import AzureBackend
from .datasets import DownloadedAzureMLDataStore, MLDataStore
from .logger import AzureMlLogger, MetricsLogger


class AzureInteractiveTrainingContext:
    """Azure interactive context that enables an interactive session with Azure ML

    Args:
        experiment_name (str): [description]
        workspace_name (str): [description]
        subscription_id (str): [description]
        resource_group (str): [description]
        dataset_dependencies (List[str]): [description]
        force_download (bool, optional): [description]. Defaults to False.

    Attributes:
        logger (MetricsLogger): a Logger
        seed: Your seed
        datasets (MLDataStore): datasets

    Raises:
        AbstractBackend.BackendException: [description]

    Example:
        You can do this:

        >>> ctx = AzureInteractiveTrainingContext(
        ...     experiment_name="",
        ...     workspace_name="",
        ...     subscription_id="",
        ...     resource_group="",
        ...     dataset_dependencies=[])
        >>> logger = ctx.logger
        >>> seed = ctx.seed
        >>> datasets = ctx.datasets

    """

    logger: MetricsLogger
    seed: int
    datasets: MLDataStore

    def __init__(
        self,
        experiment_name: str,
        workspace_name: str,
        subscription_id: str,
        resource_group: str,
        dataset_dependencies: List[str],
        force_download: bool = False,
    ):
        """[summary]

        Args:
            experiment_name (str): [description]
            workspace_name (str): [description]
            subscription_id (str): [description]
            resource_group (str): [description]
            dataset_dependencies (List[str]): [description]
            force_download (bool, optional): [description]. Defaults to False.

        Raises:
            AbstractBackend.BackendException: [description]
        """
        backend = AzureBackend()

        project_meta = {
            "workspace_name": workspace_name,
            "subscription_id": subscription_id,
            "resource_group": resource_group,
        }
        az_workspace = backend.get_workspace(project_meta)

        az_experiment = Experiment(workspace=az_workspace, name=experiment_name)

        datasets_parsed = [dataset.split(":") for dataset in dataset_dependencies]

        model_class = Mock()
        model_class.data_folder_path = "data"

        datasets = DownloadedAzureMLDataStore.from_model(
            model=model_class,
            datasets=datasets_parsed,
            workspace=az_workspace,
            force_download=force_download,
        )

        try:
            self.az_run = az_experiment.start_logging(snapshot_directory=None)
        except azureml._common.exceptions.AzureMLException as e:
            raise self.backend.parse_azureml_exception(e)

        logger = AzureMlLogger(self.az_run)

        seed = random.randint(0, 10**9)

        tags = {
            "seed": seed,
            PACKAGE_NAME: PACKAGE_VERSION,
            "python-version": PYTHON_VERSION,
            "datasets": dataset_dependencies,
        }

        self.az_run.set_tags(tags)

        self.datasets = datasets
        self.logger = logger
        self.seed = seed

    def stop(self):
        """
        Stops the interactive session with AzureML.
        Takes snapshot and marks the experiment as completed.
        """
        if self.az_run is not None:
            if click.confirm(
                "Do you want to end the job? (Remember to save file before confirming)"
            ):
                files = FileMatcher(".", include=["*.py", "*.ipynb"])
                files = [(name, name) for name in files]
                with temporary_folder(files) as temp_path:
                    self.az_run.take_snapshot(temp_path)

                self.az_run.complete()
                self.az_run = None
                self.logger.run = None
        else:
            print("Context already stopped")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Azure specific implementations
"""
from energinetml.azure.backend import AzureBackend
from energinetml.azure.interactive import AzureInteractiveTrainingContext



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

from typing import TYPE_CHECKING

import azureml

if TYPE_CHECKING:
    from energinetml.core.model import Model


class AzureSubmitContext:
    """[summary]"""

    class SubmitError(Exception):
        """[summary]"""

        pass

    class FailedToWait(SubmitError):
        """[summary]"""

        pass

    class FailedToDownload(SubmitError):
        """[summary]"""

        pass

    def __init__(self, model: "Model", az_run: azureml.core.Run):
        """[summary]

        Args:
            model (Model): [description]
            az_run (azureml.core.Run): [description]
        """
        self.model = model
        self.az_run = az_run

    def wait_for_completion(self):
        """[summary]

        Raises:
            AzureSubmitContext.FailedToWait: [description]
        """
        try:
            self.az_run.wait_for_completion(show_output=True)
        except azureml.exceptions._azureml_exception.ActivityFailedException as e:
            raise self.FailedToWait(e.message)

    def download_files(self):
        """Wrapper function for Azure Run-class download_files-function."""
        self.az_run.download_files(output_directory=self.model.path)



from unittest.mock import Mock, patch

from click.testing import CliRunner

from energinetml.cli.model.train import train
from energinetml.core.model import TrainedModel


class TestModelTrain:
    @patch("energinetml.cli.model.train.backend")
    def test__model_not_implemented__should_abort(self, backend_mock, model_path):
        """
        :param Mock backend_mock:
        :param str model_path:
        """
        runner = CliRunner()

        context = Mock()
        context.train_model.side_effect = NotImplementedError
        context.get_tags.return_value = {}
        context.get_parameters.return_value = {}
        backend_mock.get_local_training_context.return_value = context

        # Act
        result = runner.invoke(cli=train, args=["--path", model_path])

        # Assert
        assert result.exit_code == 1
        assert isinstance(result.exception, NotImplementedError)

    @patch("energinetml.cli.model.train.backend")
    def test__model_returned_something_other_than_trained_model__should_abort(
        self, backend_mock, model_path
    ):
        """
        :param Mock backend_mock:
        :param str model_path:
        """
        runner = CliRunner()

        context = Mock()
        trained_model = Mock()
        context.train_model.return_value = trained_model
        context.get_tags.return_value = {}
        context.get_parameters.return_value = {}
        backend_mock.get_local_training_context.return_value = context

        # Act
        result = runner.invoke(cli=train, args=["--path", model_path])

        # Assert
        assert result.exit_code == 1
        assert isinstance(result.exception, SystemExit)

    @patch("energinetml.cli.model.train.backend")
    @patch("energinetml.cli.model.train.TrainedModel", new=Mock)
    def test__trained_model_does_not_verify__should_abort(
        self, backend_mock, model_path
    ):
        """
        :param Mock backend_mock:
        :param str model_path:
        """
        runner = CliRunner()

        trained_model = Mock()
        trained_model.Invalid = TrainedModel.Invalid
        trained_model.verify.side_effect = TrainedModel.Invalid

        context = Mock()
        context.train_model.return_value = trained_model
        context.get_tags.return_value = {}
        context.get_parameters.return_value = {}
        backend_mock.get_local_training_context.return_value = context

        # Act
        result = runner.invoke(cli=train, args=["--path", model_path])

        # Assert
        assert result.exit_code == 1
        assert isinstance(result.exception, SystemExit)

    @patch("energinetml.core.model.Model.dump")
    @patch("energinetml.cli.model.train.backend")
    @patch("energinetml.cli.model.train.TrainedModel", new=Mock)
    def test__should_dump_trained_model_correctly(
        self, backend_mock, model_dump_mock, model_path
    ):
        """
        :param Mock backend_mock:
        :param str model_path:
        """
        runner = CliRunner()

        trained_model = Mock()

        context = Mock()
        context.train_model.return_value = trained_model
        context.get_tags.return_value = {}
        context.get_parameters.return_value = {}
        backend_mock.get_local_training_context.return_value = context

        # Act
        result = runner.invoke(cli=train, args=["--path", model_path])

        # Assert
        assert result.exit_code == 0

        model_dump_mock.assert_called_once()
        context.save_artifacts.assert_called_once()
        context.save_meta_data.assert_called_once()
